/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
# Copyright (c) 2000-2001 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Stephen Suryaputra
#                 September 1999
#This material is based upon work supported by the Defense Advanced Research
#Projects Agency Information Technology Office, (NGI Multicast Applications 
#and Architecture (NMAA-ADDON), ARPA Order No. H645, Program Code No. 9A20, 
#Issued by DARPA/CMD under Contract No. MDA972-99-C-0022.
*/

#include "swap.h"
#include "swap_analyzer.h"
#include "swap_agent.h"
#include "swap_entry.h"

static class SwapClass : public TclClass {
public:
	SwapClass() : TclClass("Agent/SWAP") {}
	TclObject* create(int, const char*const*) {
		return (new SwapAgent());
	}
} class_swap;

static char *s[] = {
	"SETUP",
	"SETUP_CONFLICT",
	"SETUP_FAIL",
	"SETUP_EGRESS",
	"COMMIT",
	"COMMIT_OK",
	"TEAR_DOWN",
};

SwapAgent::SwapAgent() : Agent(PT_SWAP_SETUP)
{
	bind("packetSize_", &size_); 
	bind("off_ip_", &off_ip_);
}

void 
SwapAgent::recv(Packet* pkt, Handler*)
{
	hdr_ip* iph = (hdr_ip *) pkt->access(off_ip_);
	//nsaddr_t src = iph->src() >> 8;
//Kamath: 02/07/2000: The data type of the value returned by iph->src()
//has been changed.
        nsaddr_t src = iph->src().addr_ >> 8;
//EOC   



	SwapMessage* swap = SwapMessage::access(pkt);
	if (swap->type < SWAP_SETUP && swap->type > SWAP_TEARDOWN) {
		analyzer_->printf("Unknown SWAP message of type %d",
				       swap->type);
		return;
	}

	FlowEntry *entry = analyzer_->lookup(swap);
	if (entry == NULL) {
		analyzer_->printf("Unrecognized flow");
		return;
	}

	analyzer_->numOfSwapPackets()++;
	analyzer_->printf("Receive %s (%s) from %d", s[swap->type],
			  entry->flowStr(), src);
	dispatch(pkt, swap, entry);
	Packet::free(pkt);
}

void
SwapAgent::dispatch(Packet *pkt, SwapMessage *swap, FlowEntry *entry)
{
	InBranch *branch;

	switch (swap->type) {
	case SWAP_SETUP:
#ifdef EGRESS
		recvSetup(entry, swap->lambda_set, swap->egress);
#else
		recvSetup(entry, swap->lambda_set);
#endif
		break;
	case SWAP_SETUP_CONFLICT:
		branch = entry->searchBranch(pkt);
		recvSetupConflict(branch);
		break;
	case SWAP_SETUP_FAIL:
		branch = entry->searchBranch(pkt);
		recvSetupFail(branch);
		break;
	case SWAP_COMMIT:
		branch = entry->searchBranch(pkt);
		recvCommit(branch, swap->lambda);
		break;
	case SWAP_COMMIT_OK:
#ifdef EGRESS
		recvCommitOK(entry, swap->egress);
#else
		recvCommitOK(entry);
#endif

		break;
#ifdef EGRESS
	case SWAP_SETUP_EGRESS:
		branch = entry->searchBranch(pkt);
		recvSetupEgress(branch);
		break;
#endif

	case SWAP_TEARDOWN:
		branch = entry->searchBranch(pkt);
		recvTeardown(branch);
		break;
	default:
		assert(false);
	}
}

#ifdef EGRESS
void
SwapAgent::recvSetup(FlowEntry *entry, bool *lambda_set, nsaddr_t egress)
#else
void
SwapAgent::recvSetup(FlowEntry *entry, bool *lambda_set)
#endif
{
	bool new_lambda_set[max_num_of_wavelength];
	bool conflict = true;
	bool path_terminated = true;
	bool has_upstream_neighbor = false;
	bool eligible_branch = false;
	bool full_path_construction = false;
	int picked_lambda= -1;
	assert(entry != 0);


/*	analyzer_->printf("SwapAgent: RecvSetup: Received lambdas:");
	for (int i = 0; i < analyzer_->numOfWavelength(); ++i) {
		if(lambda_set[i] == FALSE)
			printf("%d,", i);
	}
*/

	if (analyzer_->allocateMergingDevice(entry) == false) {
		LinkTable *l = &analyzer_->linkTable()[entry->linkIndex()];
		analyzer_->printf("SwapAgent: recvSetup: Not enough merging devices\n");
		l->agent->sendSetupConflict(entry);
		return;
	} 
	entry->outLambda() = -1;
	InBranch *branch = entry->branch();
	while (branch != 0) {


		if (! branch->hasUpstreamNeighbor()) {
			branch = branch->next();
			continue;
		}

		has_upstream_neighbor = true;
		if (branch->pktCnt() < analyzer_->highThreshPktCnt()) {
			branch = branch->next();
			continue;
		}
		// there is a high throughput branch
		eligible_branch = true;



		if (branch->isLocked()) {
			branch = branch->next();
			analyzer_->printf("SwapAgent: RecvSetup: Branch locked\n");
			continue;
		}

		conflict = false;
		bool *branch_lambda_set = branch->lambdaSet();
		LinkTable *l = &analyzer_->linkTable()[branch->linkIndex()];

		//analyzer_->printf("SwapAgent: RecvSetup: Sending lambdas:");
		for (int i = 0; i < analyzer_->numOfWavelength(); ++i) {
			new_lambda_set[i] =
				lambda_set[i] | branch_lambda_set[i];
		//	if(new_lambda_set[i] == FALSE)
		//		printf("%d,", i);

			if (new_lambda_set[i] == 0)
				full_path_construction = true;
			if (lambda_set[i] == 0)
				picked_lambda = i;
		}

		if (full_path_construction) {
			entry->outLambda() = -1; // outgoing not known yet
			branch->newState(IN_WAIT_COMMIT);
			entry->newState(OUT_WAIT_ALL_COMMIT);
			analyzer_->lockLink(branch->linkIndex());
#ifdef EGRESS
			l->agent->sendSetup(entry, new_lambda_set, egress);
#else		
			l->agent->sendSetup(entry, new_lambda_set);
#endif
			path_terminated = false;
			branch = branch->next();
			continue;
		}
		// full path construction cannot be made, do
		// wavelength conversion if possible
		entry->outLambda() = picked_lambda;
		if (analyzer_->wavelengthConvertEnable() &&
		    branch->hasAvailableLambda() &&
		    analyzer_->allocateWavelengthConverter(entry)) {
			branch->newState(IN_WAIT_COMMIT);
			entry->newState(OUT_WAIT_ALL_COMMIT);

			analyzer_->lockLink(branch->linkIndex());
#ifdef EGRESS	
			l->agent->sendSetup(entry, branch_lambda_set, egress);
#else
			l->agent->sendSetup(entry, branch_lambda_set);
#endif

			path_terminated = false;
		}
		branch = branch->next();
	}

	if (path_terminated) {
		LinkTable *l = &analyzer_->linkTable()[entry->linkIndex()];
		if (has_upstream_neighbor == false ||
		    eligible_branch == false) {
			int i = 0;
			for (; i < analyzer_->numOfWavelength(); ++i) {
				if (lambda_set[i] == 0)
					break;
			}
			if (i == analyzer_->numOfWavelength())
			{
				assert(FALSE);
			}
			entry->outLambda() = i;
			analyzer_->allocateWavelengthConverter(entry);
			if (entry->converter() == 0 &&
			    entry->mergingDevice() == 0 )
			    {
			    	printf("SwapAgent: recvSetup: Not enough wlcs/merging devices\n");
				conflict = true;
			} else {

//Kamath: 08/01/2000
//Egress switching
#ifdef EGRESS
			LinkTable *l = &analyzer_->linkTable()[entry->linkIndex()];
			for (int count = 0; count < (analyzer_->numOfWavelength()); count++)
			{
//				analyzer_->printf("SwapAgent: recvSetup: Wavelength: %d Egress: %d \n", count,egress);
//				analyzer_->printf("SwapAgent: recvSetup: Egress node: %d\n", l->lambda_entry[count].egressNode());

				if(l->lambda_entry[count].egressNode() == egress)
				{
//Needs to be changed - protocol modification - the downstream nodes should time
//out 
//					analyzer_->printf("SwapAgent: recvSetup: Found an egress switched path: Egress: %d\n", egress);	
					
					entry->outLambda() = count;
					int fc = l->lambda_entry[count].flowcount;
					l->lambda_entry[count].entry[fc]= entry;
					l->lambda_entry[count].flowcount++;
					if( l->lambda_entry[count].flowcount > max_flows_on_egress)
						analyzer_->printf("SwapAgent: recvSetup: WARNING: Number of flows exceeded storage limits. System crash likely.\n");
					
					analyzer_->deallocateWavelengthConverter(entry);	
					analyzer_->deallocateMergingDevice(entry);
//Use the wavelength converter being used by the existing egress switched path.
					entry->converter() = l->lambda_entry[count].entry[0]->converter();
					entry->newState(OUT_ACTIVE_SWITCHED);
					l->agent->sendSetupEgress(entry);
					return;
				}
			}

//					analyzer_->printf("SwapAgent: recvSetup: No egress switched path: Egress: %d Proceeding with normal setup\n", egress);	
#endif

				// normal termination
				entry->newState(OUT_COMMIT_SENT);
				l->agent->sendCommit(entry);
				return;
			}
		}
		if (conflict) {
			analyzer_->printf("SwapAgent: recvSetup: Conflict: Sending SetupConflict\n");
			l->agent->sendSetupConflict(entry);
//Kamath: 06/22/2000	
//Added to deallocate wlc in case of conflict.		
			analyzer_->deallocateWavelengthConverter(entry);
		} else if (full_path_construction == false) {
			// path terminated because full_path_construction
			// is not possible
			if (analyzer_->partialPathAllow()) {
				assert(picked_lambda != -1);
				entry->outLambda() = picked_lambda;
				entry->newState(OUT_COMMIT_SENT);
				l->agent->sendCommit(entry);
			} else {
				// cannot do conversion and partial path
				// is not allowed, setup failed
				l->agent->sendSetupFail(entry);
			}
		}
	}
}
	
void
SwapAgent::recvSetupConflict(InBranch *branch)
{
	assert(branch->state() == IN_WAIT_COMMIT ||
	       branch->state() == IN_SETUP_SENT);
	FlowEntry *entry = branch->entry();
	FlowAnalyzer *analyzer = entry->analyzer();
	analyzer->numOfSetupConflicts()++;
	LinkTable *il = &analyzer->linkTable()[branch->linkIndex()];
	analyzer->unlockLink(branch->linkIndex());
	if (branch->state() != IN_SETUP_SENT) {
		branch->newState(IN_PASSIVE);
		LinkTable *ol = &analyzer->linkTable()[entry->linkIndex()];
		assert(ol->agent != 0);
		ol->agent->sendSetupConflict(entry);
	} else
	{
//		analyzer->printf("SwapAgent: recvSetupConflict: Backing off.\n");
		branch->backoff(false);

		}
}

void
SwapAgent::recvSetupFail(InBranch *branch)
{
	assert(branch->state() == IN_WAIT_COMMIT ||
	       branch->state() == IN_SETUP_SENT);
	FlowEntry *entry = branch->entry();
	FlowAnalyzer *analyzer = entry->analyzer();
	analyzer->numOfSetupFails()++;
	LinkTable *il = &analyzer->linkTable()[branch->linkIndex()];
	analyzer->unlockLink(branch->linkIndex());
	if (branch->state() != IN_SETUP_SENT) {
		branch->newState(IN_PASSIVE);
		LinkTable *ol = &analyzer->linkTable()[entry->linkIndex()];
		ol->agent->sendSetupFail(entry);
	} else {
//Kamath: 06/15/2000
//Found that wavelength converters were not being deallocated properly.
		analyzer->deallocateWavelengthConverter(entry);
		branch->pktCnt() = 0;
		branch->newState(IN_ACTIVE);
		branch->startTimer(analyzer->flowDetectionTime());
	}
}

void
SwapAgent::recvCommit(InBranch *branch, int lambda)
{
	assert(branch->state() == IN_WAIT_COMMIT ||
	       branch->state() == IN_SETUP_SENT);
	FlowEntry *entry = branch->entry();
	FlowAnalyzer *analyzer = entry->analyzer();

//analyzer->printf("SwapAgent: recvCommit: Selected wavelength: %d\n", lambda);

	LinkTable *il = &analyzer->linkTable()[branch->linkIndex()];
	analyzer->unlockLink(branch->linkIndex());
	analyzer->allocLambda(branch->linkIndex(), lambda);
	branch->inLambda() = lambda;
	if (branch->state() == IN_WAIT_COMMIT) {
		branch->newState(IN_WAIT_OUT_SWITCHED);
		LinkTable *ol = &analyzer->linkTable()[entry->linkIndex()];
		ol->agent->sendCommit(entry);
	} else {
		// SETUP_SENT
		branch->newState(IN_SWITCHED);
//Kamath: 06/15/2000
//Found that wavelength converters were not being deallocated properly.
		analyzer->deallocateWavelengthConverter(entry);
		if (entry->state() == OUT_ACTIVE_SWITCHED)
		{
			// shifting the monitoring role upstream
			if (entry->outLambda() == lambda)
				analyzer->deallocateWavelengthConverter(entry);
			entry->newState(OUT_SWITCHED);
			entry->pktCnt() = 0;
			entry->stopTimer();
		}
		analyzer->redirect(branch);

#ifdef EGRESS
		il->agent->sendCommitOK(entry, -1);
#else	
		il->agent->sendCommitOK(entry);
#endif

}

}

#ifdef EGRESS
void
SwapAgent::recvCommitOK(FlowEntry *entry, nsaddr_t egress)
#else
void
SwapAgent::recvCommitOK(FlowEntry *entry)
#endif
{
	assert(entry->state() == OUT_WAIT_COMMIT_OK ||
	       entry->state() == OUT_COMMIT_SENT);
	FlowAnalyzer *analyzer = entry->analyzer();
	InBranch *b = entry->branch();
	if (entry->state() == OUT_WAIT_COMMIT_OK) {
		entry->newState(OUT_SWITCHED);
		while (b != 0) {
			if (b->state() == IN_WAIT_OUT_SWITCHED) {
				LinkTable *l =
				    &analyzer->linkTable()[b->linkIndex()];
#ifdef EGRESS
				l->agent->sendCommitOK(entry, egress);
#else
				l->agent->sendCommitOK(entry);
#endif

				b->newState(IN_SWITCHED);
				analyzer->redirect(b);
			} else{
				b->activate();
				}
			b = b->next();
		}
	} else {

//Kamath: 08/01/2000
//Egress switching
		LinkTable *l = &analyzer->linkTable()[entry->linkIndex()];
		
#ifdef EGRESS
		
		l->lambda_entry[entry->outLambda()].egressNode() = egress; 
		l->lambda_entry[entry->outLambda()].entry[0] = entry; 
		l->lambda_entry[entry->outLambda()].flowcount++;
//		analyzer->printf("SwapAgent: recvCommitOK: Setting lambda %d to egress %d Link table: %d Link index: %d\n", entry->outLambda(), egress, l , entry->linkIndex());	
#endif
		entry->newState(OUT_ACTIVE_SWITCHED);
		entry->pktCnt() = 0;
	

	entry->startTimer(analyzer->flowActiveTime());


	if (entry->mergingDevice() != 0)
			entry->mergingDevice()->mergeTo(entry->outLambda());
		while (b != 0) {
			b->activate();
			b = b->next();
		}
	}






}

void 
SwapAgent::recvTeardown(InBranch *branch)
{
	// Teardown is only valid for switched branch
	assert(branch->state() == IN_SWITCHED);
	analyzer_->deallocLambda(branch->linkIndex(), branch->inLambda());

	// is there any downstream neighbors
	FlowEntry *entry = branch->entry();
	if (! entry->hasDownstreamNeighbor()) {
		// no downtream, do the monitoring
		branch->activate();
		analyzer_->deallocateWavelengthConverter(entry);
		analyzer_->deallocateMergingDevice(entry);
		return;
	}
	// check other branches
	InBranch *other_branch = entry->branch();
	while(other_branch != 0) {
		if (other_branch != branch) {
			if (other_branch->state() == IN_SWITCHED) {
				// hitting merging point
				branch->activate();
				return;
			}
		}
		other_branch = other_branch->next();
	}
	branch->newState(IN_PASSIVE);
	// put all branches into passive mode
	other_branch = entry->branch();
	while (other_branch != 0) {
		if (other_branch != branch)
			{
			other_branch->deactivate();
			}
		other_branch = other_branch->next();
	}



	entry->newState(OUT_NOT_SWITCHED);
	analyzer_->deallocateWavelengthConverter(entry);
	analyzer_->deallocateMergingDevice(entry);
	LinkTable *ol = &analyzer_->linkTable()[entry->linkIndex()];
	ol->agent->sendTeardown(entry);
}

void
SwapAgent::sendSetupConflict(FlowEntry *entry)
{
	if (entry->hasOutstandingBranch())
		return;
	entry->newState(OUT_NOT_SWITCHED);
	entry->analyzer()->deallocateWavelengthConverter(entry);
	entry->analyzer()->deallocateMergingDevice(entry);
	Packet* pkt = allocpkt();
	SwapMessage *swap = SwapMessage::access(pkt);
	entry->constructSwapMessage(swap);

	hdr_cmn* ch = hdr_cmn::access(pkt);
	ch->ptype() = PT_SWAP_SETUP_CONFLICT;
	type_=PT_SWAP_SETUP_CONFLICT;
	swap->type = SWAP_SETUP_CONFLICT;
	send(entry, pkt, 0);
}

void
SwapAgent::sendSetupFail(FlowEntry *entry)
{
	if (entry->hasOutstandingBranch())
		return;
	entry->newState(OUT_NOT_SWITCHED);
	entry->analyzer()->deallocateWavelengthConverter(entry);
	entry->analyzer()->deallocateMergingDevice(entry);
	Packet* pkt = allocpkt();
	SwapMessage *swap = SwapMessage::access(pkt);
	entry->constructSwapMessage(swap);
	swap->type = SWAP_SETUP_FAIL;
	hdr_cmn* ch = hdr_cmn::access(pkt);
	type_ = PT_SWAP_SETUP_FAIL;
	ch->ptype() = PT_SWAP_SETUP_FAIL;
	send(entry, pkt, 0);
}

void
SwapAgent::sendCommit(FlowEntry *entry)
{
	assert(entry->state() == OUT_WAIT_ALL_COMMIT ||
	       entry->state() == OUT_COMMIT_SENT);
	if (entry->state() == OUT_WAIT_ALL_COMMIT) {
		if (entry->hasOutstandingBranch())
			return;
	        else
			entry->newState(OUT_WAIT_COMMIT_OK);
	} else
	{
		assert(entry->hasOutstandingBranch() == false);
		}
	if (entry->outLambda() == -1) {
		// Find a brach with the highest pkt cnt
		InBranch *highest_branch = 0;
		InBranch *cur_branch = entry->branch();
		while (cur_branch != 0) {
			if (cur_branch->state() == IN_WAIT_OUT_SWITCHED) {
				if (highest_branch == 0 ||
				    highest_branch->pktCnt() <
				           cur_branch->pktCnt())
					highest_branch = cur_branch;
			}
			cur_branch = cur_branch->next();
		}
		assert(highest_branch != 0);
		entry->outLambda() = highest_branch->inLambda();
	}
	if (entry->state() == OUT_COMMIT_SENT &&
		entry->converter() == 0 && entry->mergingDevice() == 0)
		{
		assert(false);
		}
	Packet* pkt = allocpkt();
	SwapMessage *swap = SwapMessage::access(pkt);
	entry->constructSwapMessage(swap);
	swap->type = SWAP_COMMIT;
	hdr_cmn* ch = hdr_cmn::access(pkt);
	ch->ptype() = PT_SWAP_COMMIT;
	type_ = PT_SWAP_COMMIT;
	swap->lambda = entry->outLambda();
	send(entry, pkt, 0);
}

#ifdef EGRESS
void
SwapAgent::sendCommitOK(FlowEntry *entry, nsaddr_t egress)
#else
void
SwapAgent::sendCommitOK(FlowEntry *entry)
#endif
{
	Packet* pkt = allocpkt();
	SwapMessage *swap = SwapMessage::access(pkt);
	entry->constructSwapMessage(swap);
	hdr_cmn* ch = hdr_cmn::access(pkt);
	ch->ptype() = PT_SWAP_COMMIT_OK;

	if(entry->outLambda() != 0)
		ch->uid() = entry->outLambda();
	else if((entry->branch())->inLambda() !=0)
		ch->uid() = (entry->branch())->inLambda();

	type_ = PT_SWAP_COMMIT_OK;
	swap->type = SWAP_COMMIT_OK;

#ifdef EGRESS
	if(egress == -1)
	{
//		printf("SwapAgent: sendCommitOK: src: %d dst: %d: Egress: %d \n", swap->ns_src, swap->ns_dst, here_.addr_);	
		swap->egress = here_.addr_;
	}
	else
		swap->egress = egress;
#endif
	send(entry, pkt, 0);
}

#ifndef EGRESS
void
SwapAgent::sendSetup(FlowEntry *entry, bool *lambda_set)
#else
void
SwapAgent::sendSetup(FlowEntry *entry, bool *lambda_set, nsaddr_t egress)
#endif
{
	Packet* pkt = allocpkt();
	SwapMessage *swap = SwapMessage::access(pkt);
	entry->constructSwapMessage(swap);

//Kamath: 08/01/2000
//Egress switching
#ifdef EGRESS
	if(egress == -1)
	{
//		analyzer_->printf("SwapAgent: SendSetup: Initiating setup: \n");	
//		analyzer_->printf("SwapAgent: SendSetup: src: %d dst: %d: Egress: %d \n", swap->ns_src, swap->ns_dst, here_.addr_);	
		swap->egress = here_.addr_;
	}
	else
		swap->egress = egress;

#endif

//	analyzer_->printf("SwapAgent: SendSetup: Sending wavelength set:");
	for (int i = 0; i < analyzer_->numOfWavelength(); ++i)
	{
		swap->lambda_set[i] = lambda_set[i];
//		if (lambda_set[i] == FALSE)
//			printf("%d,", i);
	}
//	printf("\n");

	swap->type = SWAP_SETUP;
	hdr_cmn* ch = hdr_cmn::access(pkt);
	ch->ptype() = PT_SWAP_SETUP;
	type_ = PT_SWAP_SETUP;
	send(entry, pkt, 0);
}

void
SwapAgent::sendTeardown(FlowEntry *entry)
{
	Packet* pkt = allocpkt();
	SwapMessage *swap = SwapMessage::access(pkt);
	entry->constructSwapMessage(swap);
	hdr_cmn* ch = hdr_cmn::access(pkt);
	ch->ptype() = PT_SWAP_TEARDOWN;

	if(entry->outLambda() != 0)
		ch->uid() = entry->outLambda();
	else if((entry->branch())->inLambda() !=0)
		ch->uid() = (entry->branch())->inLambda();

	type_ = PT_SWAP_TEARDOWN;
	swap->type = SWAP_TEARDOWN;
	send(entry, pkt, 0);
}

int 
SwapAgent::command(int argc, const char*const* argv)
{
	Tcl& tcl = Tcl::instance();

	if (argc == 3) {
		if (strcmp(argv[1], "analyzer") == 0) {
			analyzer_ = (FlowAnalyzer *) tcl.lookup(argv[2]);
			return (TCL_OK);
			//return (analyzer_ == NULL ? TCL_ERR: TCL_OK);
		}
	}
	
	return (Agent::command(argc, argv));
}

void
SwapAgent::send(FlowEntry *entry, Packet* p, Handler* h)
{
	SwapMessage *swap = SwapMessage::access(p);
	Agent::send(p, h);
}



#ifdef EGRESS

void
SwapAgent::sendSetupEgress(FlowEntry *entry)
{

	Packet* pkt = allocpkt();
	SwapMessage *swap = SwapMessage::access(pkt);
	entry->constructSwapMessage(swap);
	swap->type = SWAP_SETUP_EGRESS;
	hdr_cmn* ch = hdr_cmn::access(pkt);
	type_ = PT_SWAP_SETUP_EGRESS;
	ch->ptype() = PT_SWAP_SETUP_EGRESS;
	send(entry, pkt, 0);
}

void
SwapAgent::recvSetupEgress(InBranch *branch)
{

	assert(branch->state() == IN_WAIT_COMMIT ||
	       branch->state() == IN_SETUP_SENT);
	FlowEntry *entry = branch->entry();
	FlowAnalyzer *analyzer = entry->analyzer();
	LinkTable *il = &analyzer->linkTable()[branch->linkIndex()];
	

//Check this: Might need to be removed
	//analyzer->unlockLink(branch->linkIndex());

//Kamath: 06/15/2000
//Found that wavelength converters were not being deallocated properly.
	analyzer->deallocateWavelengthConverter(entry);
	entry->pktCnt() = 0;

	if (branch->state() == IN_WAIT_COMMIT) {
		LinkTable *ol = &analyzer->linkTable()[entry->linkIndex()];
		ol->agent->sendSetupEgress(entry);
		}

	branch->newState(IN_PASSIVE);
	entry->newState(OUT_NOT_SWITCHED);

}



#endif




