/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
# Copyright (c) 2000-2001 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Stephen Suryaputra
#                 September 1999

#This material is based upon work supported by the Defense Advanced Research
#Projects Agency Information Technology Office, (NGI Multicast Applications 
#and Architecture (NMAA-ADDON), ARPA Order No. H645, Program Code No. 9A20, 
#Issued by DARPA/CMD under Contract No. MDA972-99-C-0022.
*/

#include "net-interface.h"
#include "swap_analyzer.h"
#include "swap_address.h"


int FlowAnalyzer:: routedPktCnt = 0;

void
StatisticsTimer::expire(Event *e)
{
	int num_of_branch = analyzer_->numOfBranch();
	int wavelengthUsage = 0;
	for (int i = 0; i < num_of_branch; ++i) {
		LinkTable *link = &(analyzer_->linkTable()[i]);
		if (link->agent != NULL)
			wavelengthUsage +=
				analyzer_->numOfWavelength() -
				link->num_of_available_wavelength;
	}
	analyzer_->sumOfWavelengthUsage() += wavelengthUsage;
	analyzer_->statsSampleCnt()++;
	resched(1);
}

void
GarbageCollectTimer::expire(Event *e)
{

	for (int i = 0; i < flow_table_bucket_size; ++i) {
		FlowEntry *entry = analyzer_->flow_table_[i];
		while (entry != 0) {
			if (entry->state() != OUT_NOT_SWITCHED) {
				entry = entry->next();
				continue;
			}
			InBranch *branch = entry->branch();
			while (branch != 0) {
				if (branch->state() == IN_PASSIVE) {
					if (branch->pktCnt() == 0 &&
					    branch->garbageCnt() == 0) {
						entry->removeBranch(branch);
						InBranch *next_branch =
							branch->next();
						delete branch;
						branch = next_branch;
						continue;
					}
//Kamath:: 05/29/2000
//The resetting of the packet count occured even when the branch was switched.
//causing problems in the path setup.
				//	branch->pktCnt() = 0;
					branch->garbageCnt()--;
				}
				branch = branch->next();
			}
			if (entry->branch() == 0) {
				FlowEntry *cur_entry = entry;
				entry = entry->next();
				analyzer_->removeEntry(cur_entry);
				delete cur_entry;
				continue;
			}
			entry = entry->next();
		}
	}
	resched(garbage_collect_period);
}

void
FlowAnalyzer::printf(char *fmt, ...)
{
	va_list ap;
	Tcl& tcl = Tcl::instance();
	
	va_start(ap, fmt);
	tcl.eval("[Simulator instance] now");
//	::printf("(%s:%s)", node_str_, tcl.result());
	::printf("(%s:%lf)", node_str_, Scheduler::instance().clock() - Scheduler::instance().start());
	vprintf(fmt, ap);
	::printf("\n");
	va_end(ap);
	fflush(stdout);
}


void
FlowAnalyzer::warn_printf(char *fmt, ...)
{
	va_list ap;

	if (no_warn_) return;

	Tcl& tcl = Tcl::instance();
	va_start(ap, fmt);
	tcl.eval("[Simulator instance] now");
//	::printf("(%s:%s) WARNING: ", node_, tcl.result());
	vprintf(fmt, ap);
//	::printf("\n");
	va_end(ap);
}

void
FlowAnalyzer::printFlowTable()
{
	FlowEntry *fep;
	for (int i = 0; i < flow_table_bucket_size; ++i) {
		fep = flow_table_[i];
		while (fep != NULL) {
			printf("%s", fep->flowStr());
			fep = fep->next();
		}
	}
}

int
FlowAnalyzer::command(int argc, const char*const* argv)
{
	Tcl& tcl = Tcl::instance();

	switch(argc) {
	case 2:
		if (strcmp(argv[1], "dump-stats") == 0) {
			dumpStatistics();
			return TCL_OK;
		}
		break;
	case 3:
		if (strcmp(argv[1], "warning") == 0) {
			no_warn_ = ! atoi(argv[2]);
			return TCL_OK;
		}
		if (strcmp(argv[1], "wavelength-converter") == 0) {
			wavelength_converter_[num_of_wavelength_converter_] =
				(FlowConverter *) tcl.lookup(argv[2]);
			num_of_wavelength_converter_++;
			return TCL_OK;
		}
		if (strcmp(argv[1], "merging-device") == 0) {
			merging_device_[num_of_merging_device_] =
				(MergingDevice *) tcl.lookup(argv[2]);
			num_of_merging_device_++;
			return TCL_OK;
		}
		if (strcmp(argv[1], "flow_detect_time") == 0) {
			flow_detection_time_ = atof(argv[2]);
			return TCL_OK;
		}
		if (strcmp(argv[1], "flow_delete_time") == 0) {
			flow_deletion_time_ = atof(argv[2]);
			return TCL_OK;
		}
		if (strcmp(argv[1], "flow_active_time") == 0) {
			flow_active_time_ = atof(argv[2]);
			return TCL_OK;
		}
		if (strcmp(argv[1], "high-threshold") == 0) {
			high_thresh_pkt_cnt_ = atoi(argv[2]);
			return TCL_OK;
		}
		if (strcmp(argv[1], "low-threshold") == 0) {
			low_thresh_pkt_cnt_ = atoi(argv[2]);
			return TCL_OK;
		}
		if (strcmp(argv[1], "backoff-limit") == 0) {
			backoff_limit_ = atoi(argv[2]);
			return TCL_OK;
		}
		if (strcmp(argv[1], "backoff-period") == 0) {
			char *s;
			backoff_period_ = strtod(argv[2], &s);
			if (*s != 0) {
				if (strcmp(s, "ms") == 0)
					backoff_period_ /= 1000;
				else
				{
					assert(false);
				}
			}
			return TCL_OK;
		}
		if (strcmp(argv[1], "node") == 0) {
			strncpy(node_, argv[2], max_tcl_string);
			return TCL_OK;
		}
		if (strcmp(argv[1], "node-str") == 0) {
			strncpy(node_str_, argv[2], max_tcl_string);
			return TCL_OK;
		}
		if (strcmp(argv[1], "num-of-wavelength") == 0) {
			num_of_wavelength_ = atoi(argv[2]);
			for (int i = 0; i < max_num_of_branch; ++i)
			{
				link_table_[i].num_of_available_wavelength =
					num_of_wavelength_ - 1;
//Kamath: 08/01/200
//Egress switching
#ifdef EGRESS
				link_table_[i].lambda_entry = new LambdaEntry[num_of_wavelength_];
#endif

			}
			return TCL_OK;
		}
		if (strcmp(argv[1], "no-agent") == 0) {
			NetworkInterface *iface =
				(NetworkInterface *) tcl.lookup(argv[2]);
#ifndef NDEBUG
			for (int i = 0; i < num_of_branch_; ++i)
			{
				assert(link_table_[i].iface_num !=
				       iface->label());
			}	
#endif
			assert(num_of_branch_ < max_num_of_branch);
			LinkTable *l = &link_table_[num_of_branch_];
			++num_of_branch_;
			l->iface_num = iface->label();
			l->agent = 0;
			l->locked = false;
 			l->alloc[0] = true;
 			for (int i = 1; i < max_num_of_wavelength; ++i)
 				l->alloc[i] = false;
 			return TCL_OK;
 		}
 		break;
 	case 4:
 		if (strcmp(argv[1], "add-agent") == 0) {
 			NetworkInterface *iface =
 				(NetworkInterface *) tcl.lookup(argv[2]);
 			SwapAgent *agent = (SwapAgent *) tcl.lookup(argv[3]);
#ifndef NDEBUG
			for (int i = 0; i < num_of_branch_; ++i)
			{
				assert(link_table_[i].iface_num !=
				       iface->label());
}
#endif
 			assert(num_of_branch_ < max_num_of_branch);
 			LinkTable *l = &link_table_[num_of_branch_];
 			++num_of_branch_;
 			l->iface_num = iface->label();
 			l->agent = agent;
 			l->locked = false;
 			l->alloc[0] = true;
 			for (int i = 1; i < max_num_of_wavelength; ++i)
 				l->alloc[i] = false;
 			return TCL_OK;
 		}
 		break;
	}
	return Connector::command(argc, argv);
}

void
FlowAnalyzer::recv(Packet* p, Handler* h)
{


	// start the analyzer timer, to collect stats
 	if (timer_ == NULL) {
 		timer_ = new StatisticsTimer(this);
 		timer_->sched(1);
 	}

	if (garbage_timer_ == 0) {
		garbage_timer_ = new GarbageCollectTimer(this);
		garbage_timer_->sched(garbage_collect_period);
	}



	hdr_cmn* ch= (hdr_cmn*)p->access(hdr_cmn::offset_);
	

 	InBranch*branch = lookupAndCreate(p);
	
 	branch->pktCnt()++;
	branch->garbageCnt() = garbage_collect_cycle;
 	FlowEntry *entry = branch->entry();
	entry->pktCnt()++;

 	if (! branch->hasUpstreamNeighbor())
 		++total_pkt_received_cnt_;
//	printf("Recv: %s packet id: %d Size: %d packet cnt: %d\n", entry->flowStr(),ch->uid(), ch->size(),  entry->pktCnt());


//Kamath: 06/13/2000
//Changed the X/Y classifier model. A flow is setup when the high threshold is 
//exceeded. The previous model set up the flows only when the timer expirsed.
	if ((branch->state() == IN_ACTIVE) && (branch->pktCnt() >= highThreshPktCnt()))
	{
		branch->stopTimer();
		branch->timer_->expire(NULL);
	}



                // this is the packet entering the
 	        // system (after losses)
 	if (entry->state() == OUT_ACTIVE_SWITCHED ||
		entry->state() == OUT_ABOUT_TO_TEARDOWN || (
		entry->state() == OUT_SWITCHED && branch->state() != IN_SWITCHED)) {
		
 	 
		++total_pkt_switched_cnt_;
		if (entry->converter() == 0 && entry->mergingDevice() == 0 )
		{
			assert(false);
		}
		if (entry->converter() != 0)
			{
//Kamath: 06/13/2000
//Increment the hocpount.
			if(branch->hasUpstreamNeighbor())
				ch->switchedhopcount++;
			else
			{
				ch->switchedhopcount =0;
				ch->hopcount= 0;
				ch->iphopcount = 0;
			}
//Kamath: 08/01/2000
//Egress switching
#ifdef EGRESS
			LinkTable *l = &link_table_[entry->linkIndex()];
			l->lambda_entry[entry->outLambda()].pktCnt()++;
#endif

			entry->converter()->recv(p, h);
			
			}
		else
			{
			entry->mergingDevice()->recv(p, h);
			}
 	} else
		{

//Kamath: 06/13/2000
//Changes for hopcount statistic calculations.
			if(branch->hasUpstreamNeighbor())
				ch->iphopcount++;
			else
			{
				ch->switchedhopcount =0;
				ch->hopcount= 0;
				ch->iphopcount = 0;
			}
			send(p, h);
		}


}

FlowEntry *
FlowAnalyzer::lookup(SwapAddress *src, SwapAddress *dst)
{
 	int buck = getBucketIndex(src->val(), dst->val());
 	FlowEntry *entry = flow_table_[buck];

 	while (entry != NULL) {

 		if (entry->src()->match(src) && entry->dst()->match(dst))
 			return (entry);
 		entry = entry->next();
 	}
 	return (entry);
}

InBranch *
FlowAnalyzer::lookupAndCreate(Packet *p)
{
 	SwapAddress *src = getSrcAddr(p);
	SwapAddress *dst = getDstAddr(p);

 	FlowEntry *entry = lookup(src, dst);
 	if (entry == NULL) {
 		entry = new FlowEntry(this, src, dst);
		numFlows++;
 		insertEntry(entry);
		if (! entry->hasDownstreamNeighbor())
			entry->outLambda() = 0;
 	} else {
 		delete src;
 		delete dst;
 	}
 	InBranch *branch = entry->searchBranch(p);
 	if (branch == NULL) {
 		int i = 0;
 		hdr_cmn *cmn = hdr_cmn::access(p);
 		for (; i < num_of_branch_; ++i)
 			if (link_table_[i].iface_num == cmn->iface())
 				break;
 		assert(i != num_of_branch_);
 		branch = new InBranch(entry, i);
 		entry->insertBranch(branch);
		if (branch->state() == IN_ACTIVE)
			branch->startTimer(flow_detection_time_);
 	}
 	return branch;
}

FlowEntry *
FlowAnalyzer::lookup(SwapMessage *swap)
{
 	SwapAddress *src = getSrcAddr(swap);
 	SwapAddress *dst = getDstAddr(swap);
 	return (lookup(src, dst));
}

void
FlowAnalyzer::dumpStatistics()
{
 	::printf("Node %s Flow Statistics:\n", node_str_);
 	::printf("\tTotal Number of Signaling Packet= %d\n", num_of_swap_pkt_);
 	::printf("\tTotal Number of Transit Packet Received= %ld\n",
 		 total_pkt_received_cnt_);
 	::printf("\tTotal Number of Transit Packet Switched= %ld\n",
 		 total_pkt_switched_cnt_);
 	double x = sum_of_wavelength_usage_;
 	x /= num_of_branch_ * stats_sample_cnt_;
 	::printf("\tAverage Wavelength Usage= %.2f\n", x);
 	::printf("\tMaximum Wavelength Usage= %.2f\n",
 		 (1.0 * max_wavelength_usage_));




	int totalBackoffs = 0;
	double partialSum = 0;
	for(int count =0; count < max_num_backoffs; count++)
	{
		totalBackoffs+=backoffcount[count];
		partialSum+=(count *backoffcount[count]);
	}
	double meanBackoffCount = 0;
	if(totalBackoffs != 0)
		meanBackoffCount = partialSum/totalBackoffs;
 	::printf("\tMean number of backoffs= %.2f\n", meanBackoffCount);
	fflush(stdout);

 	::printf("\tNumber of setup conflicts= %d\n", num_of_setupconflicts);
 	::printf("\tNumber of setup failures= %d\n", num_of_setupfails);
}

void
FlowAnalyzer::lockLink(int link_idx)
{
 	assert(link_idx >= 0 && link_idx < max_num_of_branch);
 	assert(link_table_[link_idx].locked == false);
 	link_table_[link_idx].locked = true;
}

void
FlowAnalyzer::unlockLink(int link_idx)
{
 	assert(link_idx >= 0 && link_idx < max_num_of_branch);
 	assert(link_table_[link_idx].locked == true);
 	link_table_[link_idx].locked = false;
}

bool
FlowAnalyzer::allocateWavelengthConverter(FlowEntry *entry)
{
	if (entry->converter() != 0) {
		assert(false);
	}
	if (entry->src()->type() == any_address) {
		entry->converter() = 0;
		return (true);
	}
	assert(entry->outLambda() != -1);
	for (int i = 0; i < num_of_wavelength_converter_; ++i) {
		FlowConverter *wcp = wavelength_converter_[i];
		if (wcp->allocated() == false) {
			entry->converter() = wcp;
			wcp->allocated() = true;
			wcp->convertTo(entry->outLambda());
			wcp->setTarget(target_);
			return true;
		}
	}
	return (false);
}

void
FlowAnalyzer::deallocateWavelengthConverter(FlowEntry *entry)
{
	if (entry->src()->type() == any_address)
		return;
	FlowConverter *wcp = entry->converter();
	if (wcp == 0)
		return;
	wcp->allocated() = false;
	entry->converter() = 0;
}

bool
FlowAnalyzer::allocateMergingDevice(FlowEntry *entry)
{
	if (entry->src()->type() != any_address) {
		entry->mergingDevice() = 0;
		return (true);
	}
	if (entry->mergingDevice() != 0) {
		assert(false);
	}
	for (int i = 0; i < num_of_merging_device_; ++i) {
		MergingDevice *wcp = merging_device_[i];
		if (wcp->allocated() == false) {
			entry->mergingDevice() = wcp;
			wcp->allocated() = true;
			wcp->setTarget(target_);
			return true;
		}
	}
	return (false);
}

void
FlowAnalyzer::deallocateMergingDevice(FlowEntry *entry)
{
	if (entry->src()->type() != any_address)
		return;
	MergingDevice *wcp = entry->mergingDevice();
	if (wcp == 0)
		return;
	wcp->allocated() = false;
	entry->mergingDevice() = 0;
}


void 
FlowAnalyzer::allocLambda(int link_idx, int lambda)
{
	assert(link_idx >= 0 && link_idx < max_num_of_branch);
	assert(lambda >= 0 && lambda < num_of_wavelength_);
	LinkTable *l = &link_table_[link_idx];
	assert(l->alloc[lambda] == 0);
	l->alloc[lambda] = 1;
	l->num_of_available_wavelength--;

	assert(l->num_of_available_wavelength >= 0);
	int usage = num_of_wavelength_ - l->num_of_available_wavelength;

	if (usage > max_wavelength_usage_)
		max_wavelength_usage_ = usage;
}

void
FlowAnalyzer::deallocLambda(int link_idx, int lambda)
{
	assert(link_idx >= 0 && link_idx < max_num_of_branch);
	assert(lambda >= 0 && lambda < num_of_wavelength_);
	assert(link_table_[link_idx].alloc[lambda] == 1);
	LinkTable *l = &link_table_[link_idx];
	l->alloc[lambda] = 0;
	l->num_of_available_wavelength++;
	assert(l->num_of_available_wavelength <= num_of_wavelength_);
}

void
FlowAnalyzer::redirect(InBranch *branch)
{
	char link[max_tcl_string];
	char classifier[max_tcl_string];
	FlowEntry *entry = branch->entry();
	FlowAnalyzer *analyzer = entry->analyzer();
	LinkTable *il = &analyzer->linkTable()[branch->linkIndex()];
	Tcl& tcl = Tcl::instance();


	char out_dmux[max_tcl_string];
	tcl.evalf("%s set out_dmux_", node_);
	strcpy(out_dmux, tcl.result());



	tcl.evalf("%s slot %d", out_dmux, getDstDomain(entry));
	
	strncpy(link, tcl.result(), max_tcl_string);
	const char *target = link;

	if (entry->converter() != 0) {
		target = entry->converter()->name();
		entry->converter()->convertTo(entry->outLambda());
	}
	if (entry->mergingDevice() != 0) {
		tcl.evalf("%s target %s", entry->mergingDevice()->name(),
			  target);
		target = entry->mergingDevice()->name();
		entry->mergingDevice()->mergeTo(entry->outLambda());
	}
	tcl.evalf("%s set classifier_", node_);
	strncpy(classifier, tcl.result(), max_tcl_string);
	int x = il->iface_num;

//Kamath: Changed fro 16 bit shift to 8 bit shift: 16 bits caused a
//problem with lardge networks ( larger number of interfaces)
	x = x << 8;
//	x = x << 16;
	x |= branch->inLambda();
	tcl.evalf("%s install %d %s", classifier, x, target);

}

void
FlowAnalyzer::insertEntry(FlowEntry *entry)
{
	int buck = entry->buck();
	if (flow_table_[buck] != 0) {
		entry->next() = flow_table_[buck];
		flow_table_[buck]->prev() = entry;
	}
	flow_table_[buck] = entry;
}

void
FlowAnalyzer::removeEntry(FlowEntry *entry)
{
	FlowEntry *next = entry->next();
	if (next != 0)
		next->prev() = entry->prev();
	if (entry->prev() == 0) {
		flow_table_[entry->buck()] = next;
		if (next != 0)
			next->prev() = 0;
	} else
		entry->prev()->next() = next;
}

//
// Fine Granularity Flow Analyzer
//

SwapAddress *
FineFlowAnalyzer::getSrcAddr(Packet *p)
{
	hdr_ip *ns_ip = (hdr_ip *)p->access(off_ip_);
	//nsaddr_t ns_src = ns_ip->src() >> 8;

//Kamath: 02/07/2000: The data type of the value returned by iph->src()
//has been changed.
        nsaddr_t ns_src = ns_ip->src().addr_ ;
//EOC


	struct ip *ip = (struct ip *) p->accessdata();
	struct tcphdr *tcp_hdr = (struct tcphdr *)
		(p->accessdata() + sizeof(struct ip));
	return (new FineAddress(ns_src, ip->ip_src, tcp_hdr->th_sport));
}

SwapAddress *
FineFlowAnalyzer::getDstAddr(Packet *p)
{
	hdr_ip *ns_ip = (hdr_ip *)p->access(off_ip_);
//	nsaddr_t ns_dst = ns_ip->dst() >> 8;
//Kamath: 02/07/2000: The data type of the value returned by iph->src()
//has been changed.
        nsaddr_t ns_dst = ns_ip->dst().addr_;
//EOC
	struct ip *ip = (struct ip *) p->accessdata();
	struct tcphdr *tcp_hdr = (struct tcphdr *)
		(p->accessdata() + sizeof(struct ip));
	SwapAddress *a = new FineAddress(ns_dst, ip->ip_dst, tcp_hdr->th_dport);
	a->setAsDestination();
	return (a);
}

SwapAddress *
FineFlowAnalyzer::getSrcAddr(SwapMessage *swap)
{
	return (new FineAddress(swap->ns_src, swap->ip_src, swap->tcp_sport));
}

SwapAddress *
FineFlowAnalyzer::getDstAddr(SwapMessage *swap)
{
	SwapAddress *a = new FineAddress(swap->ns_dst, swap->ip_dst,
				      swap->tcp_dport);
	a->setAsDestination();
	return (a);
}

nsaddr_t
FineFlowAnalyzer::getDstDomain(FlowEntry *entry)
{
	return ((FineAddress *)entry->dst())->nsAddr();
}

nsaddr_t
FineFlowAnalyzer::getSrcDomain(FlowEntry *entry)
{
	return ((FineAddress *)entry->src())->nsAddr();
}

// Medium Flow Analyzer

SwapAddress *
MediumFlowAnalyzer::getSrcAddr(Packet *p)
{
	hdr_ip *ns_ip = (hdr_ip *)p->access(off_ip_);
//	nsaddr_t ns_src = ns_ip->src() >> 8;
//Kamath: 02/07/2000: The data type of the value returned by iph->src()
//has been changed.
        nsaddr_t ns_src = ns_ip->src().addr_ ;
///EOC
	struct ip *ip = (struct ip *) p->accessdata();
	return (new MediumAddress(ns_src, ip->ip_src));
}

SwapAddress *
MediumFlowAnalyzer::getDstAddr(Packet *p)
{
	hdr_ip *ns_ip = (hdr_ip *)p->access(off_ip_);
//	nsaddr_t ns_dst = ns_ip->dst() >> 8;

//Kamath: 02/07/2000: The data type of the value returned by iph->src()
//has been changed.
        nsaddr_t ns_dst = ns_ip->dst().addr_ ;
//EOC
	struct ip *ip = (struct ip *) p->accessdata();
	SwapAddress *a = new MediumAddress(ns_dst, ip->ip_dst);
	a->setAsDestination();
	return (a);
}

SwapAddress *
MediumFlowAnalyzer::getSrcAddr(SwapMessage *swap)
{
	return (new MediumAddress(swap->ns_src, swap->ip_src));
}

SwapAddress *
MediumFlowAnalyzer::getDstAddr(SwapMessage *swap)
{
	SwapAddress *a = new MediumAddress(swap->ns_dst, swap->ip_dst);
	a->setAsDestination();
	return (a);
}

nsaddr_t
MediumFlowAnalyzer::getDstDomain(FlowEntry *entry)
{
	return ((MediumAddress *)entry->dst())->nsAddr();
}

nsaddr_t
MediumFlowAnalyzer::getSrcDomain(FlowEntry *entry)
{
	return ((MediumAddress *)entry->src())->nsAddr();
}


// Coarse Flow Analyzer

SwapAddress *
CoarseFlowAnalyzer::getSrcAddr(Packet *p)
{
	hdr_ip *ns_ip = (hdr_ip *)p->access(off_ip_);
	//nsaddr_t ns_src = ns_ip->src() >> 8;
//Kamath: 02/07/2000: The data type of the value returned by iph->src()
//has been changed.
        nsaddr_t ns_src = ns_ip->src().addr_ ;
//EOC



	return (new CoarseAddress(ns_src));
}

SwapAddress *
CoarseFlowAnalyzer::getDstAddr(Packet *p)
{
	hdr_ip *ns_ip = (hdr_ip *)p->access(off_ip_);
//	nsaddr_t ns_dst = ns_ip->dst() >> 8;
//Kamath: 02/07/2000: The data type of the value returned by iph->src()
//has been changed.
        nsaddr_t ns_dst = ns_ip->dst().addr_ ;
//EOC



	SwapAddress *a = new CoarseAddress(ns_dst);
	a->setAsDestination();
	return (a);
}

SwapAddress *
CoarseFlowAnalyzer::getSrcAddr(SwapMessage *swap)
{
	return (new CoarseAddress(swap->ns_src));
}

SwapAddress *
CoarseFlowAnalyzer::getDstAddr(SwapMessage *swap)
{
	SwapAddress *a = new CoarseAddress(swap->ns_dst);
	a->setAsDestination();
	return (a);
}

nsaddr_t
CoarseFlowAnalyzer::getDstDomain(FlowEntry *entry)
{
	return ((CoarseAddress *)entry->dst())->nsAddr();
}

nsaddr_t
CoarseFlowAnalyzer::getSrcDomain(FlowEntry *entry)
{
	return ((CoarseAddress *)entry->src())->nsAddr();
}

// Merged Coarse Flow Analyzer

SwapAddress *
MergeCoarseFlowAnalyzer::getSrcAddr(Packet *p)
{
	return (new SwapAddress());
}

SwapAddress *
MergeCoarseFlowAnalyzer::getSrcAddr(SwapMessage *swap)
{
	return (new SwapAddress());
}

nsaddr_t
MergeCoarseFlowAnalyzer::getSrcDomain(FlowEntry *entry)
{
	assert(false);
}

// Merged Fine Flow Analyzer

SwapAddress *
MergeFineFlowAnalyzer::getSrcAddr(Packet *p)
{
	return (new SwapAddress());
}

SwapAddress *
MergeFineFlowAnalyzer::getSrcAddr(SwapMessage *swap)
{
	return (new SwapAddress());
}

nsaddr_t
MergeFineFlowAnalyzer::getSrcDomain(FlowEntry *entry)
{
	assert(false);
}

// Tcl Object Interface

static class FineFlowAnalyzerClass : public TclClass {
public:
	FineFlowAnalyzerClass() : TclClass("FlowAnalyzer/Fine") {}
	TclObject* create(int, const char*const*) {
		return (new FineFlowAnalyzer);
	}
} fine_flow_analyzer_class;

static class MediumFlowAnalyzerClass : public TclClass {
public:
	MediumFlowAnalyzerClass() : TclClass("FlowAnalyzer/Medium") {}
	TclObject* create(int, const char*const*) {
		return (new MediumFlowAnalyzer);
	}
} medium_flow_analyzer_class;

static class CoarseFlowAnalyzerClass : public TclClass {
public:
	CoarseFlowAnalyzerClass() : TclClass("FlowAnalyzer/Coarse") {}
	TclObject* create(int, const char*const*) {
		return (new CoarseFlowAnalyzer);
	}
} coarse_flow_analyzer_class;

static class MergeCoarseFlowAnalyzerClass : public TclClass {
public:
	MergeCoarseFlowAnalyzerClass() : TclClass("FlowAnalyzer/Coarse/Merge")
		{}
	TclObject* create(int, const char*const*) {
		return (new MergeCoarseFlowAnalyzer);
	}
} merge_coarse_analyzer_class;

static class MergeFineFlowAnalyzerClass : public TclClass {
public:
	MergeFineFlowAnalyzerClass() : TclClass("FlowAnalyzer/Fine/Merge")
		{}
	TclObject* create(int, const char*const*) {
		return (new MergeFineFlowAnalyzer);
	}
} merge_fine_analyzer_class;
