/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
# Copyright (c) 2000-2001 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Stephen Suryaputra
#                 September 1999

#This material is based upon work supported by the Defense Advanced Research
#Projects Agency Information Technology Office, (NGI Multicast Applications 
#and Architecture (NMAA-ADDON), ARPA Order No. H645, Program Code No. 9A20, 
#Issued by DARPA/CMD under Contract No. MDA972-99-C-0022.
*/

#ifndef SWAP_ANALYZER_H
#define SWAP_ANALYZER_H

#include <stdarg.h>
#include "packet.h"
#include "ip.h"
#include "connector.h"
#include "timer-handler.h"
#include "swap.h"
#include "swap_entry.h"
#include "swap_agent.h"
#include "flow_converter.h"
#include "merging_device.h"

extern char *status_str[];

class StatisticsTimer : public TimerHandler {
public:
	StatisticsTimer(FlowAnalyzer *analyzer) : TimerHandler() {
		analyzer_ = analyzer;
	}
protected:
        virtual void expire(Event *e);
	FlowAnalyzer *analyzer_;
};

class GarbageCollectTimer : public TimerHandler {
public:
	GarbageCollectTimer(FlowAnalyzer *analyzer) : TimerHandler() {
		analyzer_ = analyzer;
	}
protected:
        virtual void expire(Event *e);
	FlowAnalyzer *analyzer_;
};


struct LinkTable {
	int       iface_num;
	SwapAgent *agent;
	int       num_of_available_wavelength;
	LambdaEntry *lambda_entry;
	bool      alloc[max_num_of_wavelength];
  bool      locked;
#ifdef SETUP_RELEASE

//Utham Kamath: 10/13/2000
//Implementation of SETUP_READY. Defined a wavelength state to enable 
//the analyzer to decide which wavelengths to choose when sending 
//SETUP upstream. This could be used to replace "alloc" above.
	LAMBDASTATE lambdaState[max_num_of_wavelength];
#endif


};

class FlowAnalyzer : public Connector {
	friend class GarbageCollectTimer;
public:
	FlowAnalyzer() {
		numFlows = 0;
		no_warn_ = 1;
		bind("off_ip_", &off_ip_);
		num_of_wavelength_converter_ = 0;
		for (int i = 0; i < flow_table_bucket_size; ++i)
			flow_table_[i] = 0;
		for (int i = 0; i < max_num_of_wavelength_converter; ++i)
			wavelength_converter_[i] = 0;
		for (int i = 0; i < max_num_of_merging_device; ++i)
			merging_device_[i] = 0;
		for (int i = 0; i < max_num_backoffs; ++i)
			backoffcount[i] = 0;
		num_of_setupfails = 0;
		num_of_setupconflicts = 0;
		num_of_branch_ = 0;
		num_of_swap_pkt_ = 0;
		sum_of_wavelength_usage_ = 0;
		max_wavelength_usage_ = 1;
		stats_sample_cnt_ = 0;
		num_of_merging_device_ = 0;
		total_pkt_received_cnt_ = 0;
		total_pkt_switched_cnt_ = 0;
		wavelength_convert_enable_ = false;
		partial_path_allow_ = false;
		timer_ = NULL;
		garbage_timer_ = 0;
		node_str_[0] = '\0';
	}
	int& highThreshPktCnt() { return high_thresh_pkt_cnt_; }
	int& lowThreshPktCnt() { return low_thresh_pkt_cnt_; }
	double& flowDetectionTime() { return flow_detection_time_; }
	double& flowActiveTime() { return flow_active_time_; }
	double& flowDeletionTime() { return flow_deletion_time_; }
	int& numOfWavelength() { return num_of_wavelength_; }
	int& numOfSwapPackets() { return num_of_swap_pkt_; }
	int& numOfSetupFails() { return num_of_setupfails; }
	int& numOfSetupConflicts() { return num_of_setupconflicts; }
	LinkTable *linkTable() { return link_table_; }
	int& sumOfWavelengthUsage() { return sum_of_wavelength_usage_; }
	int& statsSampleCnt() { return stats_sample_cnt_; }
	int& numOfBranch() { return num_of_branch_; }
	char * nodeString() { return node_str_; }
	bool partialPathAllow() { return partial_path_allow_; }
	virtual bool wavelengthConvertEnable() {
		return wavelength_convert_enable_;
	}
	char *node() { return node_; }
	int  backoffLimit() { return backoff_limit_; }
	double backoffPeriod() { return backoff_period_; }
	
	void lockLink(int link_index);
	void unlockLink(int link_index);
	void allocLambda(int link_index, int lambda);
	void deallocLambda(int link_index, int lambda);
	void redirect(InBranch *branch);
	bool allocateWavelengthConverter(FlowEntry *entry);
	void deallocateWavelengthConverter(FlowEntry *entry);
	bool allocateMergingDevice(FlowEntry *entry);
	void deallocateMergingDevice(FlowEntry *entry);
	void printf(char *fmt, ...);
	void warn_printf(char *fmt, ...);
	void printFlowTable();
	int command(int argc, const char*const* argv);
	virtual void recv(Packet *p, Handler *h);
	InBranch *lookupAndCreate(Packet *p);
	FlowEntry *lookup(SwapMessage *swap);
	void insertEntry(FlowEntry *entry);
	void removeEntry(FlowEntry *entry);
	int getBucketIndex(unsigned int src_hash, unsigned int dst_hash) {
		return ((src_hash ^ dst_hash) % flow_table_bucket_size);
	}
	void dumpStatistics();



	int backoffcount[max_num_backoffs];


protected:
	virtual SwapAddress *getSrcAddr(Packet *p) = 0;
	virtual SwapAddress *getDstAddr(Packet *p) = 0;
	virtual SwapAddress *getSrcAddr(SwapMessage *swap) = 0;
	virtual SwapAddress *getDstAddr(SwapMessage *swap) = 0;
	virtual nsaddr_t getSrcDomain(FlowEntry *entry) = 0;
	virtual nsaddr_t getDstDomain(FlowEntry *entry) = 0;
	FlowEntry *lookup(SwapAddress *src, SwapAddress *dst);

	static int routedPktCnt;

	FILE *fout;
	SwapAgent *agent_;
	int off_ip_;
	int no_warn_;
	int high_thresh_pkt_cnt_;
	int low_thresh_pkt_cnt_;
	int backoff_limit_;
	int num_of_setupfails;
	int num_of_setupconflicts;
	double backoff_period_;
	long total_pkt_received_cnt_;
	long total_pkt_switched_cnt_;
	double flow_detection_time_;
	double flow_deletion_time_;
	double flow_active_time_;
#ifdef EGRESS
	double egresstimeout;
#endif
	char node_[max_tcl_string];
	char node_str_[max_tcl_string];
	int num_of_wavelength_converter_;
	int num_of_wavelength_;
	int num_of_swap_pkt_;
	int sum_of_wavelength_usage_;
	int stats_sample_cnt_;
	int max_wavelength_usage_;
	int num_of_branch_;
	bool wavelength_convert_enable_;
	bool partial_path_allow_;
	int num_of_merging_device_;
	int numFlows;
	LinkTable link_table_[max_num_of_branch];
	FlowConverter *wavelength_converter_[max_num_of_wavelength_converter];
	MergingDevice *merging_device_[max_num_of_merging_device];
	FlowEntry *flow_table_[flow_table_bucket_size];
	StatisticsTimer *timer_;
	GarbageCollectTimer *garbage_timer_;
};

class FineFlowAnalyzer : public FlowAnalyzer {
public:
	FineFlowAnalyzer() : FlowAnalyzer() {}
protected:
	virtual SwapAddress *getSrcAddr(Packet *p);
	virtual SwapAddress *getDstAddr(Packet *p);
	virtual SwapAddress *getSrcAddr(SwapMessage *swap);
	virtual SwapAddress *getDstAddr(SwapMessage *swap);
	virtual nsaddr_t getSrcDomain(FlowEntry *entry);
	virtual nsaddr_t getDstDomain(FlowEntry *entry);
};

class MediumFlowAnalyzer : public FlowAnalyzer {
public:
	MediumFlowAnalyzer() : FlowAnalyzer() {}
protected:
	virtual SwapAddress *getSrcAddr(Packet *p);
	virtual  SwapAddress *getDstAddr(Packet *p);
	virtual SwapAddress *getSrcAddr(SwapMessage *swap);
	virtual  SwapAddress *getDstAddr(SwapMessage *swap);
	virtual nsaddr_t getSrcDomain(FlowEntry *entry);
	virtual nsaddr_t getDstDomain(FlowEntry *entry);
};

class CoarseFlowAnalyzer : public FlowAnalyzer {
public:
	CoarseFlowAnalyzer() : FlowAnalyzer() {}
protected:
	virtual SwapAddress *getSrcAddr(Packet *p);
	virtual SwapAddress *getDstAddr(Packet *p);
	virtual SwapAddress *getSrcAddr(SwapMessage *swap);
	virtual SwapAddress *getDstAddr(SwapMessage *swap);
	virtual nsaddr_t getSrcDomain(FlowEntry *entry);
	virtual nsaddr_t getDstDomain(FlowEntry *entry);
};

class MergeCoarseFlowAnalyzer : public CoarseFlowAnalyzer {
public:
	MergeCoarseFlowAnalyzer() : CoarseFlowAnalyzer() {}
	virtual bool wavelengthConvertEnable() { return true; }
protected:
	virtual SwapAddress *getSrcAddr(Packet *p);
	virtual SwapAddress *getSrcAddr(SwapMessage *swap);
	virtual nsaddr_t getSrcDomain(FlowEntry *entry);
};

class MergeFineFlowAnalyzer : public FineFlowAnalyzer {
public:
	MergeFineFlowAnalyzer() : FineFlowAnalyzer() {}
	virtual bool wavelengthConvertEnable() { return true; }
protected:
	virtual SwapAddress *getSrcAddr(Packet *p);
	virtual SwapAddress *getSrcAddr(SwapMessage *swap);
	virtual nsaddr_t getSrcDomain(FlowEntry *entry);
};

#endif




