/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
# Copyright (c) 2000-2001 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Stephen Suryaputra
#                 September 1999

#This material is based upon work supported by the Defense Advanced Research
#Projects Agency Information Technology Office, (NGI Multicast Applications 
#and Architecture (NMAA-ADDON), ARPA Order No. H645, Program Code No. 9A20, 
#Issued by DARPA/CMD under Contract No. MDA972-99-C-0022.
*/

#include <assert.h>
#include <math.h>
#include "swap.h"
#include "swap_entry.h"
#include "swap_address.h"
#include "swap_analyzer.h"
#include "random.h"

FlowEntry::FlowEntry(FlowAnalyzer *analyzer, SwapAddress *src, SwapAddress *dst)
{
	assert(analyzer != NULL && src != NULL && dst != NULL);
	buck_ = analyzer->getBucketIndex(src->val(), dst->val());
	state_ = OUT_NOT_SWITCHED;
	pkt_cnt_ = 0;
	analyzer_ = analyzer;
	next_ = 0;
	prev_ = 0;
	out_lambda_ = -1;
	clearStats();
	branch_ = 0;
	timer_ = 0;
	src_ = src;
	dst_ = dst;
	int iface = dst_->outIface(analyzer_);
	int i = 0;
	for (; i < analyzer_->numOfBranch(); ++i)
		if (analyzer_->linkTable()[i].iface_num == iface)
			break;
	assert(i != max_num_of_branch);
	link_idx_ = i;
	if (analyzer_->linkTable()[i].agent == NULL)
		out_lambda_ = 0;
	wcp_ = 0;
	merging_device_ = 0;


}

FlowEntry::~FlowEntry() 
{ 
	assert(src_ != 0);
	assert(dst_ != 0);
	delete src_;
	delete dst_; 
	if (timer_ != 0) {
		assert(timer_->active() == false);
		delete timer_;
	}
}

char *
FlowEntry::flowStr()
{
	static char s[med_str_len+5];
	char s1[med_str_len];
	char s2[med_str_len];
	
	strncpy(s1, src_->str(), med_str_len);
	strncpy(s2, dst_->str(), med_str_len);
	snprintf(s, med_str_len +5 , "(%s->%s:%d)", s1, s2, out_lambda_);
	return (s);
}

InBranch *
FlowEntry::searchBranch(Packet *p)
{
	hdr_cmn *cmn = hdr_cmn::access(p);

	InBranch *branch = branch_;
	while (branch != NULL) {
		if( analyzer_->linkTable()[branch->linkIndex()].iface_num == cmn->iface())
			break;
		branch = branch->next();
	}

	return (branch);
}

void
FlowEntry::insertBranch(InBranch *branch)
{
	LinkTable *il = &analyzer_->linkTable()[branch->linkIndex()];
	branch->next() = branch_;
	branch_ = branch;
		
}

bool
FlowEntry::hasDownstreamNeighbor()
{
	return (analyzer_->linkTable()[link_idx_].agent != 0);
}

void
FlowEntry::newState(OutState new_state)
{
	static char *state_str_[] = {
		"OUT_NOT_SWITCHED",
		"OUT_COMMIT_SENT",
		"OUT_ACTIVE_SWITCHED",
		"OUT_ABOUT_TO_TEARDOWN",
		"OUT_WAIT_ALL_COMMIT",
		"OUT_WAIT_COMMIT_OK",
		"OUT_SWITCHED",
	};

//	analyzer_->printf("%s %s->%s", flowStr(),
		//	  state_str_[state_],
		//	  state_str_[new_state]);
	state_ = new_state;
}


void
FlowEntry::constructSwapMessage(SwapMessage *swap)
{
	src_->fillSwapMessage(swap);
	dst_->fillSwapMessage(swap);
}

void
FlowEntry::removeBranch(InBranch *branch)
{
	assert(branch != 0);
	InBranch *bcur = branch_;
	InBranch *bprev = 0;
	while (bcur != 0) {
		if (bcur == branch)
			break;
		bprev = bcur;
		bcur = bcur->next();
	}
	assert(bcur != 0);
	if (bprev == 0)
		branch_ = bcur->next();
	else
		bprev->next() = bcur->next();
}

bool
FlowEntry::hasOutstandingBranch()
{
	InBranch *b = branch_;
	while (b != 0) {
		if (b->state() == IN_WAIT_COMMIT)
			return (true);
		b = b->next();
	}
	return (false);
}

void
FlowEntry::startTimer(double expire)
{
	if (timer_ == 0) {
		timer_ = new FlowTimer(this);
	}
	timer_->sched(expire);
	assert(timer_->active() == false);
	timer_->active() = true;
}

void
FlowEntry::stopTimer()
{
	assert(timer_ != 0);
	assert(timer_->active() == true);
	timer_->cancel();
	timer_->active() = false;
}

void
FlowTimer::expire(Event *e)
{
	FlowAnalyzer *analyzer = entry_->analyzer();
//	analyzer->printf("FlowTimer: expire: %s :Pkt count: %d\n", entry_->flowStr(), entry_->pktCnt());
//Kamath: 08/01/2000
//Egress switching

	InBranch *branch = entry_->branch();
#ifdef EGRESS
	LinkTable *l = &analyzer->linkTable()[entry_->linkIndex()];
	int pktcnt = l->lambda_entry[entry_->outLambda()].pktCnt();

//	analyzer->printf("FlowTimer: expire: Packet count: Egress: %d\n", pktcnt);
//	analyzer->printf("FlowTimer: expire: Packet count: Flow: %d\n", entry_->pktCnt());
//	analyzer->printf("FlowTimer: expire: Out lambda: %d\n", entry_->outLambda());
//	analyzer->printf("FlowTimer: expire: Egress: %d\n", l->lambda_entry[entry_->outLambda()].egressNode());
//	analyzer->printf("Egress: Link table: %d Link index: %d\n",  l , entry_->linkIndex());	

#else
	int pktcnt = entry_->pktCnt();
#endif



	switch (entry_->state()) {
	case OUT_ACTIVE_SWITCHED:
		if (pktcnt < analyzer->lowThreshPktCnt()) {
			entry_->newState(OUT_ABOUT_TO_TEARDOWN);
		}
		break;
	case OUT_ABOUT_TO_TEARDOWN:
		if (pktcnt < analyzer->lowThreshPktCnt()) {
#ifdef EGRESS
//Remove the entry from the table when the path is being deleted.
//			analyzer->printf("FlowTimer: expire: Deleting the egress switched path: Egress: %d\n", l->lambda_entry[entry_->outLambda()].egressNode());
			l->lambda_entry[entry_->outLambda()].pktCnt() = 0;
			l->lambda_entry[entry_->outLambda()].entry[0] = 0;
			l->lambda_entry[entry_->outLambda()].egressNode() = -1;
#endif
			

			entry_->newState(OUT_NOT_SWITCHED);
			LinkTable *ol =
				&analyzer->linkTable()[entry_->linkIndex()];
			analyzer->deallocateWavelengthConverter(entry_);
			analyzer->deallocateMergingDevice(entry_);

//Remove entries for all entries using this wl
#ifdef EGRESS

		for(int count = 1; count < l->lambda_entry[entry_->outLambda()].flowcount; count++)
		{
			l->lambda_entry[entry_->outLambda()].entry[count]->newState(OUT_NOT_SWITCHED);
			l->lambda_entry[entry_->outLambda()].entry[count] = 0;
		}
		l->lambda_entry[entry_->outLambda()].flowcount = 0;



#endif

			InBranch *branch = entry_->branch();
			while (branch != 0) {
				branch->deactivate();
				branch = branch->next();
			}

			ol->agent->sendTeardown(entry_);




			active_ = false;
//			analyzer->printf("FlowTimer: expire: Deleting flow: %s: Pkt count: %d\n", entry_->flowStr(), entry_->pktCnt());
			return;
		} else
			entry_->newState(OUT_ACTIVE_SWITCHED);
		break;
	default:
		assert(false);
	}
	entry_->pktCnt()=0;
#ifdef EGRESS
	l->lambda_entry[entry_->outLambda()].pktCnt() = 0;
#endif

	resched(analyzer->flowActiveTime());
}

// InBranch

InBranch::InBranch(FlowEntry *entry, int link_idx)
{
	link_idx_ = link_idx;
	entry_ = entry;
	next_ = 0;
	timer_ = 0;
	pkt_cnt_ = 0;
	in_lambda_ = -1;
	backoff_count_ = 0;
		
	if (! entry_->hasDownstreamNeighbor() &&
	    entry_->analyzer()->linkTable()[link_idx_].agent != 0) {
		state_ = IN_ACTIVE;
		return;
	}

	if (entry_->analyzer()->linkTable()[link_idx_].agent != 0 &&
	    (entry_->state() == OUT_ACTIVE_SWITCHED ||
	     entry_->state() == OUT_SWITCHED)) {
		state_ = IN_ACTIVE;
		return;
	}
	state_ = IN_PASSIVE;
}

InBranch::~InBranch()
{
	if (timer_ != 0)
		delete timer_;
}

bool
InBranch::hasUpstreamNeighbor()
{
	FlowAnalyzer *analyzer = entry_->analyzer();
	return (analyzer->linkTable()[link_idx_].agent != 0);
}

bool
InBranch::isLocked()
{
	FlowAnalyzer *analyzer = entry_->analyzer();
	return (analyzer->linkTable()[link_idx_].locked);
}

bool *
InBranch::lambdaSet()
{
	FlowAnalyzer *analyzer = entry_->analyzer();
	return (analyzer->linkTable()[link_idx_].alloc);
}

bool
InBranch::hasAvailableLambda()
{
	FlowAnalyzer *analyzer = entry_->analyzer();
	return(  analyzer->linkTable()[link_idx_].num_of_available_wavelength > 0);
}

void
InBranch::newState(InState new_state)
{
	state_ = new_state;
}

void
InBranch::startTimer(double expire)
{
	if (timer_ == 0) {
		timer_ = new BranchTimer(this);
	}
	assert(timer_->active() == false);
	timer_->sched(expire);
	timer_->active() = true;
}

//Kamath: 06/13/2000
//Added a stopTimer to allow chnages to the X/Y classifier model.

void
InBranch::stopTimer()
{
	assert(timer_ != 0);
	assert(timer_->active() == true);
	timer_->cancel();
	timer_->active() = false;
}




void
InBranch::activate()
{
	pkt_cnt_ = 0;
	if (hasUpstreamNeighbor()) {
		state_ = IN_ACTIVE;
		startTimer(entry_->analyzer()->flowDetectionTime());
	} else {
		assert(state_ == IN_PASSIVE);
		if (timer_ != 0)
		{
			assert(timer_->active() == false);
			}
	}
}

void
InBranch::deactivate()
{
	if (state_ == IN_PASSIVE) {
		if (timer_ != 0)
			{
			assert(timer_->active() == false);
			}
		return;
	}
	assert(timer_->active() == true);
	
	timer_->cancel();
	state_ = IN_PASSIVE;
	pkt_cnt_ = 0;
	timer_->active() = false;
}

void
InBranch::backoff(bool in_timer)
{
	FlowAnalyzer *analyzer = entry_->analyzer();
	if (entry_->state() == OUT_NOT_SWITCHED)
		analyzer->deallocateWavelengthConverter(entry_);
	if (backoff_count_ < analyzer->backoffLimit()) {
		backoff_count_++;
		newState(IN_BACKOFF);
		double backoff_period =
			Random::uniform(analyzer->backoffPeriod() *
					(1U << backoff_count_));
//		analyzer->printf("InBranch: Backoff: %s Backoff count = %d, period = %f",
//				 entry_->flowStr(), backoff_count_, backoff_period);
		if (in_timer)
			{
			timer_->resched(backoff_period);
			}
		else
			{

			startTimer(backoff_period);
			}
	} else {
		backoff_count_ = 0;
		pkt_cnt_ = 0;
		newState(IN_ACTIVE);
//		analyzer->printf("InBRanch: Backkof: %s Not able to backoff", entry_->flowStr());
		if (in_timer)
		{
		
			timer_->resched(analyzer->flowDetectionTime());
				}
		else
		{
			startTimer(analyzer->flowDetectionTime());
				}
	}
}

void 
BranchTimer::expire(Event *e)
{

	FlowEntry *entry = branch_->entry();
	FlowAnalyzer *analyzer = entry->analyzer();
	int pkt_cnt = branch_->pktCnt();

//	analyzer->printf("BranchTimer: expire: %s packetcnt: %d\n", entry->flowStr(), pkt_cnt);
	branch_->pktCnt() = 0;
	LinkTable *l = &analyzer->linkTable()[branch_->linkIndex()];
	switch (branch_->state()) {
	case IN_PASSIVE:
		assert(false);
		break;
	case IN_ACTIVE:
//Utham Kamath: Changed the condition from > to >= 
		if (pkt_cnt >= analyzer->highThreshPktCnt() &&
		    branch_->hasAvailableLambda()) {
			branch_->backoffCount() = 0;
	case IN_BACKOFF:

//Kamath: 06/29/2000   
//The availability of wavelengths should be checked before 
//proceeding even if it is after a backoff.
		    if( !(branch_->hasAvailableLambda()))
		    { 
//				analyzer->printf("BranchTimer: expire: Not enough lambdas\n"),
				branch_->backoff(true);
				return;
	 		}

			assert(l->agent != 0);
			if (branch_->isLocked()) {
//				analyzer->printf("BranchTimer: expire: Interface %d is locked",
//						 l->iface_num);
				branch_->backoff(true);
				return;
			}
			// whether it is going to be useful or not,
			// reserve the converter first. Just in case.

				if (entry->converter() == 0 && entry->mergingDevice() == 0) 
				{
					analyzer->allocateWavelengthConverter(entry);
					analyzer->allocateMergingDevice(entry);
				}
				if (entry->converter() == 0 && entry->mergingDevice() == 0) 
				{
//					analyzer->printf("BranchTimer: expire: Not enough resources\n");
					branch_->backoff(true);
					return;
				}
//Kamath: 06/13/2000
//Statistics collection.
			analyzer->backoffcount[branch_->backoffCount()]++;


			branch_->newState(IN_SETUP_SENT);
			analyzer->lockLink(branch_->linkIndex());
//Kamath: 08/01/2000
//Egress switching
#ifdef EGRESS
			l->agent->sendSetup(entry, l->alloc, -1);
#else
			l->agent->sendSetup(entry, l->alloc);
#endif

//		analyzer->printf("BranchTimer: expire: Creating flow: %s packetcnt: %d\n", entry->flowStr(), pkt_cnt);
			active_ = false;
			return;
		}
//		analyzer->printf("BranchTimer: expire: Backing off.\n");
		resched(analyzer->flowDetectionTime());
		break;
	default:
		assert(false);
	}
}

#ifdef EGRESS

LambdaEntry::LambdaEntry()
{
	pkt_cnt_ = 0;
	egress = -1;
	flowcount =0;
}

LambdaEntry::~LambdaEntry() 
{ 
}

#endif
